#!/usr/bin/env python
# encoding: utf-8

'''
@author: Tsuyoshi Hombashi
'''

from __future__ import absolute_import
from __future__ import with_statement
import sys

try:
    import json
except ImportError:
    import simplejson as json

import six
import thutils

import tcconfig
import tcconfig.traffic_control


def parse_option():
    parser = thutils.option.ArgumentParserObject()
    parser.make(version=tcconfig.VERSION)

    group = parser.add_argument_group("Traffic Control")
    group.add_argument(
        "--device", action="append", required=True,
        help="network device name (e.g. eth0)")

    return parser.parse_args()


@thutils.main.Main
def main():
    options = parse_option()

    thutils.initialize_library(__file__, options)
    thutils.common.verify_install_command(["tc"])

    subproc_wrapper = thutils.subprocwrapper.SubprocessWrapper()
    tc_param = {}

    for device in options.device:
        tcconfig.verify_network_interface(device)

        tc = tcconfig.traffic_control.TrafficControl(
            subproc_wrapper, device)
        tc_param.update(tc.get_tc_parameter())

    six.print_(json.dumps(tc_param, indent=4))

    return 0


if __name__ == '__main__':
    sys.exit(main())



# encoding: utf-8

'''
@author: Tsuyoshi Hombashi
'''

from __future__ import absolute_import

import pytest
import six

import tcconfig.parser


@pytest.fixture
def filter_parser():
    return tcconfig.parser.TcFilterParser()


@pytest.fixture
def qdisc_parser():
    return tcconfig.parser.TcQdiscParser()


class Test_TcFilterParser_parse_filter:

    @pytest.mark.parametrize(["value", "expected"], [
        [None, []],
        ["", []],
        [
            six.b("""
            filter parent 1: protocol ip pref 1 u32
            filter parent 1: protocol ip pref 1 u32 fh 801: ht divisor 1
            filter parent 1: protocol ip pref 1 u32 fh 801::800 order 2048 key ht 801 bkt 0 flowid 1:1
              match c0a8000a/ffffffff at 16
            filter parent 1: protocol ip pref 2 u32
            filter parent 1: protocol ip pref 2 u32 fh 800: ht divisor 1
            filter parent 1: protocol ip pref 2 u32 fh 800::800 order 2048 key ht 800 bkt 0 flowid 1:2
              match 00000000/00000000 at 16
            """),
            [
                {'flowid': '1:1', 'network': '192.168.0.10/32', 'port': None},
                {'flowid': '1:2', 'network': '0.0.0.0/0', 'port': None},
            ],
        ],
        [
            six.b("""
            filter parent 1: protocol ip pref 1 u32
            filter parent 1: protocol ip pref 1 u32 fh 801: ht divisor 1
            filter parent 1: protocol ip pref 1 u32 fh 801::800 order 2048 key ht 801 bkt 0 flowid 1:1
              match c0a80000/ffffff00 at 16
              match 00000050/0000ffff at 20
            filter parent 1: protocol ip pref 2 u32
            filter parent 1: protocol ip pref 2 u32 fh 800: ht divisor 1
            filter parent 1: protocol ip pref 2 u32 fh 800::800 order 2048 key ht 800 bkt 0 flowid 1:2
              match 00000000/00000000 at 16
            """),
            [
                {'flowid': '1:1', 'network': '192.168.0.0/24', 'port': 80},
                {'flowid': '1:2', 'network': '0.0.0.0/0', 'port': None},
            ],
        ],
        [
            six.b("""
            filter parent 1: protocol ip pref 1 u32
            filter parent 1: protocol ip pref 1 u32 fh 801: ht divisor 1
            filter parent 1: protocol ip pref 1 u32 fh 801::800 order 2048 key ht 801 bkt 0 flowid 1:3
              match c0a8000a/ffffffff at 12
              match 00001f90/0000ffff at 20
            filter parent 1: protocol ip pref 2 u32
            filter parent 1: protocol ip pref 2 u32 fh 800: ht divisor 1
            filter parent 1: protocol ip pref 2 u32 fh 800::800 order 2048 key ht 800 bkt 0 flowid 1:2
              match 00000000/00000000 at 12
            """),
            [
                {'flowid': '1:3', 'network': '192.168.0.10/32', 'port': 8080},
                {'flowid': '1:2', 'network': '0.0.0.0/0', 'port': None},
            ],
        ],
    ])
    def test_normal(self, filter_parser, value, expected):
        assert filter_parser.parse_filter(value) == expected


class Test_TcFilterParser_parse_incoming_device:

    @pytest.mark.parametrize(["value", "expected"], [
        ["", None],
        [None, None],
        [
            six.b("""filter parent ffff: protocol ip pref 49152 u32
filter parent ffff: protocol ip pref 49152 u32 fh 800: ht divisor 1
filter parent ffff: protocol ip pref 49152 u32 fh 800::800 order 2048 key ht 800 bkt 0 flowid 1:
  match 00000000/00000000 at 0
        action order 1: mirred (Egress Redirect to device ifb0) stolen
        index 3795 ref 1 bind 1"""),
            "ifb0",
        ],
    ])
    def test_normal(self, filter_parser, value, expected):
        assert filter_parser.parse_incoming_device(value) == expected

    """
    @pytest.mark.parametrize(["value", "expected"], [
        [None, AttributeError],
    ])
    def test_exception(self, filter_parser, value, expected):
        with pytest.raises(expected):
            filter_parser.parse_incoming_device(value)
    """


class Test_TcQdiscParser_parse:

    @pytest.mark.parametrize(["value", "expected"], [
        ["", {}],
        [
            six.b("""qdisc prio 1: root refcnt 2 bands 3 priomap  1 2 2 2 1 2 0 0 1 1 1 1 1 1 1 1
qdisc netem 11: parent 1:1 limit 1000 delay 10.0ms loss 0.1% corrupt 0.1%
qdisc tbf 20: parent 11:1 rate 100Mbit burst 100000b limit 10000b"""),
            {
                'parent': '1:1',
                'corrupt': "0.1",
                'delay': '10.0',
                'loss': '0.1',
                'rate': '100M',
            },
        ],
        [
            six.b("""qdisc prio 1: dev eth0 root refcnt 2 bands 3 priomap  1 2 2 2 1 2 0 0 1 1 1 1 1 1 1 1
qdisc netem 11: dev eth0 parent 1:1 limit 1000 delay 10.0ms  2.0ms loss 0.1% corrupt 0.1%
qdisc tbf 20: dev eth0 parent 11:1 rate 100Mbit burst 100000b limit 10000b
qdisc pfifo_fast 0: dev ifb0 root refcnt 2 bands 3 priomap  1 2 2 2 1 2 0 0 1 1 1 1 1 1 1 1"""),
            {
                'parent': '1:1',
                'corrupt': "0.1",
                'delay': '10.0',
                'delay-distro': '2.0',
                'loss': '0.1',
                'rate': '100M',
            },
        ],
        [
            six.b("""qdisc prio 1: root refcnt 2 bands 3 priomap  1 2 2 2 1 2 0 0 1 1 1 1 1 1 1 1
qdisc netem 11: parent 1:1 limit 1000 delay 1.0ms loss 0.01%
qdisc tbf 20: parent 11:1 rate 250Kbit burst 1600b lat 268.8ms"""),
            {
                'parent': '1:1',
                'delay': '1.0',
                'loss': '0.01',
                'rate': '250K',
            },
        ]
    ])
    def test_normal(self, qdisc_parser, value, expected):
        assert qdisc_parser.parse(value) == expected

    @pytest.mark.parametrize(["value", "expected"], [
        [None, AttributeError],
    ])
    def test_exception(self, qdisc_parser, value, expected):
        with pytest.raises(expected):
            qdisc_parser.parse(value)



# encoding: utf-8

'''
@author: Tsuyoshi Hombashi
'''

import itertools
import platform

import dataproperty
import pingparsing
import pytest
import six
import thutils
import tcconfig


DEVICE = "eth0"
WAIT_TIME = 5  # [sec]


@pytest.fixture
def dst_host_option(request):
    return request.config.getoption("--dst-host")


@pytest.fixture
def dst_host_ex_option(request):
    return request.config.getoption("--dst-host-ex")


@pytest.fixture
def subproc_wrapper():
    return thutils.subprocwrapper.SubprocessWrapper()


@pytest.fixture
def pingparser():
    return pingparsing.PingParsing()


@pytest.fixture
def transmitter():
    transmitter = pingparsing.PingTransmitter()
    transmitter.ping_option = "-f -q"
    transmitter.waittime = WAIT_TIME

    return transmitter


class NormalTestValue:
    RATE_LIST = [
        "",
        "--rate 100K",
        "--rate 0.5M",
    ]
    DELAY_LIST = [
        "",
        "--delay 100",
    ]
    DELAY_DISTRO_LIST = [
        "",
        "--delay-distro 20",
    ]
    PACKET_LOSS_RATE_LIST = [
        "",
        "--loss 0.1",
    ]
    CORRUPTION_RATE_LIST = [
        "",
        "--corrupt 0.1",
    ]
    DIRECTION_LIST = [
        "",
        "--direction outgoing",
        "--direction incoming",
    ]
    NETWORK_LIST = [
        "",
        "--network 192.168.0.10",
        "--network 192.168.0.0/24",
    ]
    PORT_LIST = [
        "",
        "--port 80",
    ]
    OVERWRITE_LIST = [
        "",
        "--overwrite",
    ]


class Test_tcconfig:
    """
    Tests of in this class are inappropriate for Travis CI.
    Execute following command at the local environment  when running tests:
      python setup.py test --addopts --runxfail

    These tests are expected to execute on following environment:
       - Linux(debian) w/ iputils-ping package
       - English environment (for parsing ping output)
    """

    @pytest.mark.xfail
    @pytest.mark.parametrize(
        [
            "rate", "delay", "delay_distro", "loss", "corrupt",
            "direction", "network", "port", "overwrite",
        ],
        [
            opt_list
            for opt_list in itertools.product(
                NormalTestValue.RATE_LIST,
                NormalTestValue.DELAY_LIST,
                NormalTestValue.DELAY_DISTRO_LIST,
                NormalTestValue.PACKET_LOSS_RATE_LIST,
                NormalTestValue.CORRUPTION_RATE_LIST,
                NormalTestValue.DIRECTION_LIST,
                NormalTestValue.NETWORK_LIST,
                NormalTestValue.PORT_LIST,
                NormalTestValue.OVERWRITE_LIST)
        ])
    def test_smoke(
            self, subproc_wrapper, rate, delay, delay_distro, loss, corrupt,
            direction, network, port, overwrite):
        command = " ".join([
            "tcset",
            "--device " + DEVICE,
            rate, delay, delay_distro, loss,
            direction, network, port, overwrite,
        ])
        assert subproc_wrapper.run(command) == 0

        assert subproc_wrapper.run("tcdel --device " + DEVICE) == 0

    @pytest.mark.xfail
    @pytest.mark.parametrize(["overwrite", "expected"], [
        ["", 0],
        ["--overwrite", 255],
    ])
    def test_config_file(self, tmpdir, subproc_wrapper, overwrite, expected):
        p = tmpdir.join("tcconfig.json")
        config = """{
    "eth0": {
        "outgoing": {
            "network=192.168.0.10/32, port=8080": {
                "delay": "10.0", 
                "loss": "0.01", 
                "rate": "250K", 
                "delay-distro": "2.0"
            }, 
            "network=0.0.0.0/0": {}
        }, 
        "incoming": {
            "network=192.168.10.0/24": {
                "corrupt": "0.02", 
                "rate": "1500K"
            }, 
            "network=0.0.0.0/0": {}
        }
    }
}
"""
        p.write(config)

        subproc_wrapper.run("tcdel --device " + DEVICE)
        command = " ".join(["tcset -f ", str(p), overwrite])
        assert subproc_wrapper.run(command) == expected

        proc = subproc_wrapper.popen_command("tcshow --device " + DEVICE)
        tcshow_stdout, _stderr = proc.communicate()
        assert thutils.loader.JsonLoader.loads(
            tcshow_stdout) == thutils.loader.JsonLoader.loads(config)

        assert subproc_wrapper.run("tcdel --device " + DEVICE) == 0


class Test_tcset_one_network:
    """
    Tests of in this class are inappropriate for Travis CI.
    Execute following command at the local environment  when running tests:
      python setup.py test --addopts "--dst-host=<hostname or IP address>"

    These tests are expected to execute on following environment:
       - Linux(debian) w/ iputils-ping package
       - English environment (for parsing ping output)
    """

    @pytest.mark.parametrize(["delay"], [
        [100],
    ])
    def test_const_latency(
            self, dst_host_option, subproc_wrapper,
            transmitter, pingparser, delay):
        if dataproperty.is_empty_string(dst_host_option):
            # alternative to pytest.mark.skipif
            return

        subproc_wrapper.run("tcdel --device " + DEVICE)
        transmitter.destination_host = dst_host_option

        # w/o latency tc ---
        result = transmitter.ping()
        pingparser.parse(result)
        without_tc_rtt_avg = pingparser.rtt_avg

        # w/ latency tc ---
        command_list = [
            "tcset",
            "--device " + DEVICE,
            "--delay %d" % (delay),
        ]
        assert subproc_wrapper.run(" ".join(command_list)) == 0

        result = transmitter.ping()
        pingparser.parse(result)
        with_tc_rtt_avg = pingparser.rtt_avg

        # assertion ---
        rtt_diff = with_tc_rtt_avg - without_tc_rtt_avg
        assert rtt_diff > (delay / 2.0)

        # finalize ---
        assert subproc_wrapper.run("tcdel --device " + DEVICE) == 0

    @pytest.mark.skipif("platform.system() == 'Windows'")
    @pytest.mark.parametrize(["delay", "delay_distro"], [
        [100, 50],
    ])
    def test_const_latency_distro(
            self, dst_host_option, subproc_wrapper,
            transmitter, pingparser, delay, delay_distro):
        if dataproperty.is_empty_string(dst_host_option):
            # alternative to pytest.mark.skipif
            return

        subproc_wrapper.run("tcdel --device " + DEVICE)
        transmitter.destination_host = dst_host_option

        # w/o latency tc ---
        result = transmitter.ping()
        pingparser.parse(result)
        without_tc_rtt_avg = pingparser.rtt_avg
        without_tc_rtt_mdev = pingparser.rtt_mdev

        # w/ latency tc ---
        command_list = [
            "tcset",
            "--device " + DEVICE,
            "--delay %d" % (delay),
            "--delay-distro %d" % (delay_distro),
        ]
        assert subproc_wrapper.run(" ".join(command_list)) == 0

        result = transmitter.ping()
        pingparser.parse(result)
        with_tc_rtt_avg = pingparser.rtt_avg
        with_tc_rtt_mdev = pingparser.rtt_mdev

        # assertion ---
        rtt_diff = with_tc_rtt_avg - without_tc_rtt_avg
        assert rtt_diff > (delay / 2.0)

        rtt_diff = with_tc_rtt_mdev - without_tc_rtt_mdev
        assert rtt_diff > (delay_distro / 2.0)

        # finalize ---
        subproc_wrapper.run("tcdel --device " + DEVICE)

    @pytest.mark.parametrize(["option", "value"], [
        ["--loss", 10],
        ["--corrupt", 10],
    ])
    def test_const_packet_loss(
            self, dst_host_option, subproc_wrapper,
            transmitter, pingparser, option, value):
        if dataproperty.is_empty_string(dst_host_option):
            # alternative to pytest.mark.skipif
            return

        subproc_wrapper.run("tcdel --device " + DEVICE)
        transmitter.destination_host = dst_host_option

        # w/o packet loss tc ---
        result = transmitter.ping()
        pingparser.parse(result)
        without_tc_loss = (
            pingparser.packet_receive / float(pingparser.packet_transmit)) * 100.0

        # w/ packet loss tc ---
        command_list = [
            "tcset",
            "--device " + DEVICE,
            "%s %f" % (option, value),
        ]
        assert subproc_wrapper.run(" ".join(command_list)) == 0

        result = transmitter.ping()
        pingparser.parse(result)
        with_tc_loss = (
            pingparser.packet_receive / float(pingparser.packet_transmit)) * 100.0

        # assertion ---
        loss_diff = without_tc_loss - with_tc_loss
        assert loss_diff > (value / 2.0)

        # finalize ---
        subproc_wrapper.run("tcdel --device " + DEVICE)


class Test_tcset_two_network:
    """
    Tests of in this class are inappropriate for Travis CI.
    Execute following command at the local environment  when running tests:
      python setup.py test --addopts \
        "--dst-host=<hostname or IP address> --dst-host-ex=<hostname or IP address>"

    These tests are expected to execute on following environment:
       - Linux(debian) w/ iputils-ping package
       - English environment (for parsing ping output)
    """

    def test_network(
            self, dst_host_option, dst_host_ex_option, subproc_wrapper,
            transmitter, pingparser):
        if any([
            dataproperty.is_empty_string(dst_host_option),
            dataproperty.is_empty_string(dst_host_ex_option),
        ]):
            # alternative to pytest.mark.skipif
            return

        subproc_wrapper.run("tcdel --device " + DEVICE)
        delay = 100

        # tc to specific network ---
        command_list = [
            "tcset",
            "--device " + DEVICE,
            "--delay %d" % (delay),
            "--network " + dst_host_ex_option,
        ]
        assert subproc_wrapper.run(" ".join(command_list)) == 0

        # w/o tc network ---
        transmitter.destination_host = dst_host_option
        result = transmitter.ping()
        pingparser.parse(result)
        without_tc_rtt_avg = pingparser.rtt_avg

        # w/ tc network ---
        transmitter.destination_host = dst_host_ex_option
        result = transmitter.ping()
        pingparser.parse(result)
        with_tc_rtt_avg = pingparser.rtt_avg

        # assertion ---
        rtt_diff = with_tc_rtt_avg - without_tc_rtt_avg
        assert rtt_diff > (delay / 2.0)

        # finalize ---
        subproc_wrapper.run("tcdel --device " + DEVICE)



# encoding: utf-8

'''
@author: Tsuyoshi Hombashi
'''

VERSION = "0.6.1"


def verify_network_interface(device):
    import netifaces

    if device not in netifaces.interfaces():
        raise ValueError("invalid network interface: " + device)
